#!/usr/bin/env python

# exit codes
NOT_OK = 1
OK = 0

# modules for OPUS internal poller
import sys                  # access to command-line args
import string               # string functions
import opusutil             # utilites like PrintMsg
import re                   # regular expression handling
import traceback            # printing of stack traces 

# python modules for this application
import os
import getopt
import time

# local modules for this application
import base_poller
import siname
import hst_archdbquery

# exceptions
class UnresolvedSwitchFile(Exception):
  pass

class FailedKWSet(Exception):
  pass

class FailedSQLRead(Exception):
  pass

class UsageError(Exception):
  pass

class HaltEvent(Exception):
  pass

class DBException(Exception):
  pass


class Bestswitch(base_poller.InternalPoller):
  """
########################################################################
Class: Bestswitch

Description:
------------
The class is for interactive or pipeline setting of the best calibration
switches in the archive catalog *_ref_data tables.  The switch values
are set for any catalog entries containing NULL switch values, such as
those that have just been ingested.

Members:
--------
outpath - directory for working files
instruments - list of which instruments to process
update_limit - max number of SQL updates to do at a time (avoids
               transaction log overflow)
arch_dbobj = database object for accessing archive catalog tables

Methods:
--------
usage() - prints an interactive usage message and raises an exception

History:
-------- -----  --------  --------------------------------------
09/23/02 xxxxx  MSwam     first version
07/15/10 57271  MSwam     remove db retry parameter

########################################################################
  """
  def __init__(self):
    # initialize members
    self._outpath = None
    self._instruments = None
    self._update_limit = 5000
    self._arch_server = os.environ['ARCH_SERVER']
    self._arch_db     = os.environ['ARCH_DB']
    self._arch_dbobj  = None
    self._kw_server   = os.environ['KW_SERVER']
    self._kw_db       = os.environ['KW_DB']

  def initialize(self):
    #
    # call superclass constructor
    base_poller.InternalPoller.__init__(self)


  #--------------------------------------------------------------------------
  def usage(self):
    opusutil.PrintMsg('F','Interactive Usage: bestswitch.py [-h] [-n instrument] [-l update_limit]')
    raise UsageError

  #--------------------------------------------------------------------------
  #
  # Name: interactive
  #
  # Description: Performs command line parsing and set-up for
  #              processing in interactive mode.  Runs the
  #              task for the command-line arguments given.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #--------------------------------------------------------------------------
  def interactive(self):
    #
    # parse command line arguments
    try:
      opts, args = getopt.getopt(sys.argv[1:], "hn:l:", 
                   ["help","instrument=","update_limit="])
    except getopt.GetoptError:
      # print help information and raise exception
      self.usage()

    # parse options
    for flag, value in opts:
      if flag in ("-h", "--help"):
        self.usage()
      if flag in ("-n", "--instrument"):
        self._instruments = string.split(value)
      if flag in ("-l", "--update_limit"):
        self._update_limit = value

    # connect to db
    self._arch_dbobj = hst_archdbquery.HST_Arch_db(self._arch_server,
                                        self._arch_db)
    
    # set the directory for file operations to the current directory
    self._outpath = os.getcwd()

    # run the computations
    self.do_bestswitch()

    # close the database connection
    self._arch_dbobj.close()

    return

  #-------------------------------------------------------------------------
  # 
  # Name: pipeline_init
  #
  # Description: Initialization for pipeline mode, including reading
  #              XML file of reference file specs and resource file
  #              parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def pipeline_init(self):
    # obtain resource items in a dictionary and assign to globals
    #
    try:
      resource_items = base_poller.InternalPoller.get_resource_items(self,
                            ["OUTPATH",
                             "UPDATE_LIMIT",
                             "ARCH_SERVER",
                             "ARCH_DB",
                             "KW_SERVER",
                             "KW_DB",
                             "SUPPORTED_INSTR"] )
    except:
      opusutil.PrintMsg("E","failed to get required resource entry")
      raise
    #
    opusutil.PrintMsg("I","Resource items read.")
    self._outpath = resource_items['OUTPATH']
    self._update_limit = int(resource_items['UPDATE_LIMIT'])
    self._instruments = string.split(resource_items['SUPPORTED_INSTR'])
    self._arch_server = resource_items['ARCH_SERVER']
    self._arch_db     = resource_items['ARCH_DB']
    self._kw_server   = resource_items['KW_SERVER']
    self._kw_db       = resource_items['KW_DB']
    #
    # connect to the database
    self._arch_dbobj = hst_archdbquery.HST_Arch_db(self._arch_server,
                                        self._arch_db)
    return

  #-------------------------------------------------------------------------
  # 
  # Name: SqlUpdateFromFile
  #
  # Description: Reads a file containing SQL update code and
  #              runs the query, applying it in row-limiting "chunks",
  #              and infinitely re-trying query failures until they succeed 
  #              ("zombie").
  #
  # Inputs: filename - name of SQL file to apply to database
  #
  # Returns: None
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/19/03 xxxxx  MSwam     first version
  # 10/13/10 57271  MSwam     chunked updates dont work in MSSQL
  #
  #-------------------------------------------------------------------------
  def SqlUpdateFromFile(self, filename):
    #
    # read SQL code body from file and add to query
    querytxt = " "
    try:
      for line in open(filename, 'r').readlines():
        querytxt = querytxt + line
    except IOError:
      opusutil.PrintMsg("E","Failure reading from SQL file "+filename)
      raise FailedSQLRead

    # run the query
    self._arch_dbobj.update(querytxt)
    #
    return

  #-------------------------------------------------------------------------
  # 
  # Name: ListOfListsToFile
  #
  # Description: Writes the contents of a list of lists to a file.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def ListOfListsToFile(self, datasets, filename):
    #
    try:
      outfd = open(filename, "w")
    except IOError:
      opusutil.PrintMsg("E","failed to open output file "+filename)
      raise
    for dataset in datasets:
      outfd.write(dataset[0]+'\n')
    outfd.close()

  #-------------------------------------------------------------------------
  # 
  # Name: do_bestswitch
  #
  # Description: Compute the bestswitch values.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/23/02 xxxxx  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def do_bestswitch(self):
    #
    # the kwset tool only handles one SI at a time, so split the processing
    for instr in self._instruments:
      #
      # form the filename holding the switch names
      name = "OPUS_DEFINITIONS_DIR:"+instr+ ".switches"
      switch_file = opusutil.StretchFile(name)
      if string.find(switch_file, "OPUS_DEFINITIONS_DIR") != -1:
        # found, so stretch did not resolve
        opusutil.PrintMsg("E","Failed to resolve stretch for "+name)
        raise UnresolvedSwitchFile
        
      # get the exposure names that need processing
      datasets = self._arch_dbobj.empty_switches(instr, switch_file)
      if len(datasets) == 0:
        opusutil.PrintMsg("I","No empty switches for "+instr)
        continue
      opusutil.PrintMsg("I",str(len(datasets))+" datasets found for "+instr)
      opusutil.PrintMsg("D","datasets = "+str(datasets))

      # write exposures to a file for feeding kwset
      datasets_file = self._outpath + os.sep + 'datasets.'+instr
      self.ListOfListsToFile(datasets, datasets_file)

      # run the kwset tool on them to generate SQL output
      sql_file = self._outpath + os.sep + 'sql.'+instr
      command_str = ' '
      try:
        command_str = ('kwset -k ' + switch_file +
                       ' -of '+ sql_file + 
                       ' -f ' + datasets_file +
                       ' -ks '+ self._kw_server +
                       ' -kd '+ os.environ['KW_DB'] +
                       ' -ds '+ self._arch_server +
                       ' -dd '+ os.environ['ARCH_DB'])
        opusutil.PrintMsg("I","command_str="+command_str)
        os.system(command_str)
      except:
        opusutil.PrintMsg("E","Error running kwset for "+datasets_file)
        opusutil.PrintMsg("E","command_str="+command_str)
        raise FailedKWSet
 
      # apply the SQL output
      self.SqlUpdateFromFile(sql_file)
    #
    return

  #-------------------------------------------------------------------------
  # Name: processHaltEvent
  #
  # Description: Handle a HALT event (override base class).
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 03/01/05 51433  MSwam     use base poller class
  #
  #-------------------------------------------------------------------------
  def processHaltEvent(self, event):
    #
    opusutil.PrintMsg("I","Received HALT event. Stopping.")
    self._arch_dbobj.close()
    raise HaltEvent

#-------------------------------------------------------------------------
# The main routine. Handles interactive and pipeline mode.
#
# History:
# -------- -----  --------  --------------------------------------
# 02/28/05 51433  MSwam     use base_poller base class, and only support
#                           interactive mode for now
#-------------------------------------------------------------------------
#
# initialize
try:
  poller = Bestswitch()
  poller.initialize()
  poller.pipeline_init()
#
except base_poller.InteractiveMode:
  opusutil.PrintMsg("I","Running command-line mode.")
  try:
    poller.interactive()
  except UsageError:
    sys.exit(NOT_OK)
  except:
    traceback.print_exc()
    opusutil.PrintMsg("F","Command-line mode Failed.")
    sys.exit(NOT_OK)
  sys.exit(OK)
#
except:
  traceback.print_exc()
opusutil.PrintMsg("E","Initialization failed. "+
                      "Task only supports interactive mode for now.")
sys.exit(NOT_OK)

# UNTIL Time pollers are working for Java/Python internal pollers
# this task must be run as a external poller (pr 50377)
#
#opusutil.PrintMsg("I","Pipeline set-up complete.")
#try:
#  poller.run()
#except HaltEvent:
#  sys.exit(OK)
#except:
#  traceback.print_exc()
#  opusutil.PrintMsg("E","pipeline_process failed")
#  sys.exit(NOT_OK)
